# Generated by Django 5.2.9 on 2025-12-22 06:51

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("relations", "0003_relation_relations_r_subj_content_type_and_more"),
        ("sample_project", "0001_squashed_0011_alter_place_feature_code_and_more"),
    ]

    operations = [
        migrations.CreateModel(
            name="IsParentOf",
            fields=[
                (
                    "relation_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="relations.relation",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
            bases=("relations.relation",),
        ),
    ]



from apis_core.utils.rdf import Attribute, Filter


class E21_PersonFromDNB:
    filter_type_is_differentiated_person = Filter(
        [("rdf:type", "gndo:DifferentiatedPerson")]
    )

    forename = Attribute(
        [
            "gndo:preferredNameEntityForThePerson/gndo:forename",
            "gndo:variantNameForThePerson/gndo:forename",
            "gndo:forename",
            "gndo:preferredNameForThePerson",
        ]
    )
    surname = Attribute(
        [
            "gndo:preferredNameEntityForThePerson/gndo:surname",
            "gndo:variantNameForThePerson/gndo:surname",
            "gndo:surname",
        ]
    )
    alternative_names = Attribute("gndo:variantNameForThePerson")
    date_of_birth = Attribute("gndo:dateOfBirth")
    date_of_death = Attribute("gndo:dateOfDeath")
    same_as = Attribute("owl:sameAs")
    profession = Attribute("gndo:professionOrOccupation")


class E21_PersonFromWikidata:
    filter_p31_is_q5 = Filter([("wdt:P31", "wd:Q5")])

    forename = Attribute("wdt:P735/rdfs:label")
    surname = Attribute("wdt:P734/rdfs:label")
    date_of_birth = Attribute("wdt:P569")
    date_of_death = Attribute("wdt:P570")
    same_as = Attribute(
        ["owl:sameAs", "wdtn:P227", "wdtn:P1566", "wdtn:P214", "wdtn:P244"]
    )



import functools
import logging

from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ValidationError
from django.db import models
from django.db.models import Case, When
from django.db.models.base import ModelBase
from model_utils.managers import InheritanceManager

from apis_core.generic.abc import GenericModel

logger = logging.getLogger(__name__)


class RelationManager(InheritanceManager):
    def create_between_instances(self, subj, obj, *args, **kwargs):
        subj_object_id = subj.pk
        subj_content_type = ContentType.objects.get_for_model(subj)
        obj_object_id = obj.pk
        obj_content_type = ContentType.objects.get_for_model(obj)
        rel = self.create(
            subj_object_id=subj_object_id,
            subj_content_type=subj_content_type,
            obj_object_id=obj_object_id,
            obj_content_type=obj_content_type,
        )
        logger.debug("Created relation %s between %s and %s", rel.name(), subj, obj)
        return rel

    def to_content_type_with_targets(self, content_type):
        """
        Return the queryset annotated with the target content type
        and object id, based on the content_type that is passed.
        """
        return self.annotate(
            target_content_type=Case(
                When(subj_content_type=content_type, then="obj_content_type"),
                default="subj_content_type",
            ),
            target_id=Case(
                When(subj_content_type=content_type, then="obj_object_id"),
                default="subj_object_id",
            ),
        )


# This ModelBase is simply there to check if the needed attributes
# are set in the Relation child classes.
class RelationModelBase(ModelBase):
    def __new__(metacls, name, bases, attrs):
        if name == "Relation":
            return super().__new__(metacls, name, bases, attrs)
        else:
            new_class = super().__new__(metacls, name, bases, attrs)
            if not (new_class._meta.abstract or new_class._meta.proxy):
                if not hasattr(new_class, "subj_model"):
                    raise ValueError(
                        "%s inherits from Relation and must therefore specify subj_model"
                        % name
                    )
                if not hasattr(new_class, "obj_model"):
                    raise ValueError(
                        "%s inherits from Relation and must therefore specify obj_model"
                        % name
                    )

                # `subj_model` or `obj_model` being a list was supported in an earlier
                # version of apis, but it is not anymore
                if isinstance(getattr(new_class, "subj_model", None), list):
                    raise ValueError("%s.subj_model must not be a list" % name)
                if isinstance(getattr(new_class, "obj_model", None), list):
                    raise ValueError("%s.obj_model mut not be a list" % name)

            if not new_class._meta.ordering:
                logger.warning(
                    f"{name} inherits from Relation but does not specify 'ordering' in its Meta class. "
                    "Empty ordering could result in inconsitent results with pagination. "
                    "Set a ordering or inherit the Meta class from Relation.",
                )

            return new_class


@functools.cache
def get_by_natural_key(natural_key: str):
    app_label, name = natural_key.lower().split(".")
    return ContentType.objects.get_by_natural_key(app_label, name).model_class()


class Relation(GenericModel, models.Model, metaclass=RelationModelBase):
    subj_content_type = models.ForeignKey(
        ContentType, on_delete=models.CASCADE, related_name="relation_subj_set"
    )
    subj_object_id = models.PositiveIntegerField(null=True)
    subj = GenericForeignKey("subj_content_type", "subj_object_id")
    obj_content_type = models.ForeignKey(
        ContentType, on_delete=models.CASCADE, related_name="relation_obj_set"
    )
    obj_object_id = models.PositiveIntegerField(null=True)
    obj = GenericForeignKey("obj_content_type", "obj_object_id")

    objects = RelationManager()

    class Meta:
        indexes = [
            models.Index(
                fields=["subj_content_type"], name="relations_r_subj_content_type"
            ),
            models.Index(fields=["subj_object_id"], name="relations_r_subj_object_id"),
            models.Index(
                fields=["obj_content_type"], name="relations_r_obj_content_type"
            ),
            models.Index(fields=["obj_object_id"], name="relations_r_obj_object_id"),
            models.Index(
                fields=["subj_content_type", "subj_object_id"],
                name="relations_r_subj_c_t_o_i",
            ),
            models.Index(
                fields=["obj_content_type", "obj_object_id"],
                name="relations_r_obj_c_t_o_i",
            ),
        ]

    def save(self, *args, **kwargs):
        subj_model = getattr(self, "subj_model", None)
        if subj_model and self.subj_content_type is subj_model:
            raise ValidationError(f"{self.subj} is not of type {subj_model}")
        obj_model = getattr(self, "obj_model", None)
        if obj_model and self.obj_content_type is obj_model:
            raise ValidationError(f"{self.obj} is not of type {obj_model}")
        super().save(*args, **kwargs)

    @property
    def subj_to_obj_text(self) -> str:
        if hasattr(self, "name"):
            return f"{self.subj} {self.name()} {self.obj}"
        return f"{self.subj} relation to {self.obj}"

    @property
    def obj_to_subj_text(self) -> str:
        if hasattr(self, "reverse_name"):
            return f"{self.obj} {self.reverse_name()} {self.subj}"
        return f"{self.obj} relation to {self.subj}"

    def __str__(self):
        return self.subj_to_obj_text

    @classmethod
    def subj_model_type(cls):
        model = cls.subj_model
        return get_by_natural_key(model) if isinstance(model, str) else model

    @classmethod
    def obj_model_type(cls):
        model = cls.obj_model
        return get_by_natural_key(model) if isinstance(model, str) else model

    @classmethod
    def name(cls) -> str:
        return cls._meta.verbose_name

    @classmethod
    def reverse_name(cls) -> str:
        return cls._meta.verbose_name + " reverse"

    @classmethod
    def name_and_reverse_name(cls) -> str:
        """
        Return a string with both the name and the reverse name.

        If they are identical, return only the name.
        """
        if cls.name() != cls.reverse_name():
            return f"{cls.name()} - {cls.reverse_name()}"
        return cls.name()



from rest_framework import pagination, viewsets

from apis_core.generic.schema import GenericAutoSchema

from .filterbackends import GenericFilterBackend
from .helpers import first_member_match, makeclassprefix, module_paths
from .serializers import GenericHyperlinkedModelSerializer, serializer_factory


class InjectFacetsLimitOffsetPagination(pagination.LimitOffsetPagination):
    """
    This pagination class injects a `facets` dict into the API response object.
    It looks for a `get_facets` method in the models class and if this
    method exists it uses its return value as the value of the `facets` key in
    the API response.
    To make the API response always paginated, we have to set the `default_limit`
    attribute to something other than None.
    """

    default_limit = 20
    facets = None

    def paginate_queryset(self, queryset, request, view=None):
        if hasattr(queryset.model, "get_facets"):
            self.facets = queryset.model.get_facets(queryset)
        return super().paginate_queryset(queryset, request, view)

    def get_paginated_response(self, data):
        response = super().get_paginated_response(data)
        if self.facets is not None:
            response.data.update({"facets": self.facets})
            response.data = dict(sorted(response.data.items()))
        return response


class ModelViewSet(viewsets.ModelViewSet):
    """
    API ViewSet for a generic model.
    The queryset is overridden by the first match from
    the `first_member_match` helper.
    The serializer class is overridden by the first match from
    the `first_member_match` helper.
    """

    filter_backends = [GenericFilterBackend]
    schema = GenericAutoSchema()
    pagination_class = InjectFacetsLimitOffsetPagination

    def dispatch(self, *args, **kwargs):
        self.model = kwargs.get("contenttype").model_class()
        return super().dispatch(*args, **kwargs)

    def get_queryset(self):
        queryset_methods = module_paths(
            self.model, path="querysets", suffix="ViewSetQueryset"
        )
        queryset = first_member_match(queryset_methods) or (lambda x: x)
        return queryset(self.model.objects.all())

    def get_serializer_class(self):
        renderer = getattr(getattr(self, "request", {}), "accepted_renderer", None)
        serializer_class_modules = module_paths(
            self.model, path="serializers", suffix="Serializer"
        )
        if renderer is not None:
            prefix = makeclassprefix(renderer.format)
            serializer_class_modules = (
                module_paths(
                    self.model, path="serializers", suffix=f"{prefix}Serializer"
                )
                + serializer_class_modules
            )

        serializer_class = first_member_match(
            serializer_class_modules,
            getattr(renderer, "serializer", GenericHyperlinkedModelSerializer),
        )
        return serializer_factory(
            self.model, serializer=serializer_class, action=self.action
        )



from apis_core.utils.rdf import Attribute, Filter


class E74_GroupFromDNB:
    filter_type_1 = Filter([("rdf:type", "gndo:CorporateBody")])
    filter_type_2 = Filter([("rdf:type", "gndo:Company")])
    filter_type_3 = Filter([("rdf:type", "gndo:FictiveCorporateBody")])
    filter_type_4 = Filter([("rdf:type", "gndo:MusicalCorporateBody")])
    filter_type_5 = Filter([("rdf:type", "gndo:OrganOfCorporateBody")])
    filter_type_6 = Filter([("rdf:type", "gndo:ProjectOrProgram")])
    filter_type_7 = Filter([("rdf:type", "gndo:ReligiousAdministrativeUnit")])
    filter_type_8 = Filter([("rdf:type", "gndo:ReligiousCorporateBody")])

    label = Attribute(
        [
            "gndo:preferredNameForTheCorporateBody",
            "gndo:variantNameForTheCorporateBody",
        ]
    )
    start_date = Attribute("gndo:dateOfEstablishment")
    end_date = Attribute("gndo:dateOfTermination")
    sameas = Attribute("owl:sameAs")


class E74_GroupFromWikidata:
    filter_p31_is_q414147 = Filter([("wdt:P31", "wd:Q414147")])

    label = Attribute(["schema:about/rdfs:label", "rdfs:label"])
    same_as = Attribute(
        ["owl:sameAs", "wdtn:P227", "wdtn:P1566", "wdtn:P214", "wdtn:P244"]
    )



import functools
import logging
from collections import defaultdict

from django.apps import apps
from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.db.models import Case, CharField, F, Q, Value, When
from django.db.models.base import ModelBase
from django.db.models.functions import Concat
from django.urls import NoReverseMatch, reverse

from apis_core.apis_metainfo.models import RootObject
from apis_core.relations.models import Relation
from apis_core.utils.settings import apis_base_uri

NEXT_PREV = getattr(settings, "APIS_NEXT_PREV", True)

logger = logging.getLogger(__name__)


class AbstractEntityModelBase(ModelBase):
    def __new__(metacls, name, bases, attrs):
        if name == "AbstractEntity":
            return super().__new__(metacls, name, bases, attrs)
        else:
            new_class = super().__new__(metacls, name, bases, attrs)
            if not new_class._meta.ordering:
                logger.warning(
                    f"{name} inherits from AbstractEntity but does not specify 'ordering' in its Meta class. "
                    "Empty ordering could result in inconsitent results with pagination. "
                    "Set a ordering or inherit the Meta class from AbstractEntity.",
                )

            return new_class


class AbstractEntity(RootObject, metaclass=AbstractEntityModelBase):
    """
    Abstract super class which encapsulates common logic between the
    different entity kinds and provides various methods relating to either
    all or one specific entity kind.

    Most of the class methods are designed to be used in the subclass as they
    are considering contexts which depend on the subclass entity type.
    So they are to be understood in that dynamic context.
    """

    class Meta:
        abstract = True

    @functools.cached_property
    def get_prev_id(self):
        if NEXT_PREV:
            prev_instance = (
                type(self)
                .objects.filter(id__lt=self.id)
                .order_by("-id")
                .only("id")
                .first()
            )
            if prev_instance is not None:
                return prev_instance.id
        return False

    @functools.cached_property
    def get_next_id(self):
        if NEXT_PREV:
            next_instance = (
                type(self)
                .objects.filter(id__gt=self.id)
                .order_by("id")
                .only("id")
                .first()
            )
            if next_instance is not None:
                return next_instance.id
        return False

    def import_data(self, data):
        super().import_data(data)
        if "same_as" in data:
            self._uris = data["same_as"]
            self.save()
        if "relations" in data:
            self.create_relations_to_uris = data["relations"]
            self.save()

    def get_default_uri(self):
        try:
            route = reverse("GetEntityGenericRoot", kwargs={"pk": self.pk})
        except NoReverseMatch:
            route = reverse("apis_core:GetEntityGeneric", kwargs={"pk": self.pk})
        base = apis_base_uri().strip("/")
        return f"{base}{route}"

    @classmethod
    def get_facet_label(cls):
        if "db_string" in cls._meta.get_fields():
            return F("db_string")
        return Concat(
            Value(f"{cls._meta.verbose_name.title()} "),
            F("pk"),
            output_field=CharField(),
        )

    @classmethod
    def get_facets(cls, queryset):
        facets = defaultdict(dict)
        if getattr(cls, "enable_facets", False):
            my_content_type = ContentType.objects.get_for_model(queryset.model)

            # we filter all the relations that are somehow connected to instances of this queryset
            # we are only interested in the id and the type of the object the relations point to,
            # so we store those in `target_id` and `target_content_type`
            query_filter = Q(
                subj_content_type=my_content_type, subj_object_id__in=queryset
            ) | Q(
                Q(obj_content_type=my_content_type, obj_object_id__in=queryset),
                ~Q(subj_content_type=my_content_type),
            )
            rels = (
                Relation.objects.to_content_type_with_targets(my_content_type).values(
                    "target_id", "target_content_type"
                )
            ).filter(query_filter)

            # we use the ids to get a the labels of the targets
            # and we store those in an id: label dict
            target_ids = [rel["target_id"] for rel in rels]
            entity_classes = list(
                filter(lambda x: issubclass(x, AbstractEntity), apps.get_models())
            )

            # this should be moved to an EntityManager once the entites app is in apis_core
            when_clauses_classes = [
                When(
                    **{f"{cls.__name__.lower()}__isnull": False},
                    then=cls.get_facet_label(),
                )
                for cls in entity_classes
            ]
            identifiers = (
                RootObject.objects_inheritance.filter(id__in=target_ids)
                .select_subclasses()
                .annotate(label=Case(*when_clauses_classes))
                .values("pk", "label")
            )

            identifiers = {item["pk"]: item["label"] for item in identifiers}

            for rel in rels:
                _id = rel["target_id"]
                content_type = rel["target_content_type"]
                facetname = (
                    "relation_to_" + ContentType.objects.get_for_id(content_type).name
                )
                if _id in facets[facetname]:
                    facets[facetname][_id]["count"] += 1
                else:
                    facets[facetname][_id] = {
                        "count": 1,
                        "name": identifiers.get(_id, "Unknown"),
                    }
        return facets



from apis_core.utils.rdf import Attribute, Filter


class E53_PlaceFromDNB:
    filter_type_is_naturalgeographicunit = Filter(
        [("rdf:type", "gndo:NaturalGeographicUnit")]
    )
    filter_type_is_tcb_or_adu = Filter(
        [("rdf:type", "gndo:TerritorialCorporateBodyOrAdministrativeUnit")]
    )

    label = Attribute("gndo:preferredNameForThePlaceOrGeographicName")
    longitude = Attribute("""
    SELECT ?longitude
    WHERE {
      ?subject geo:hasGeometry/geo:asWKT ?point .
      BIND(REPLACE(str(?point), "Point \\\\( \\\\+?(-?\\\\d+.\\\\d+).*", "$1") as ?longitude)
    }
    """)
    latitude = Attribute("""
    SELECT ?latitude
    WHERE {
      ?subject geo:hasGeometry/geo:asWKT ?point .
      BIND(REPLACE(str(?point), "^Point\\\\s*\\\\(\\\\s*[+-]?\\\\d+\\\\.\\\\d+\\\\s+([+-]?\\\\d+\\\\.\\\\d+)\\\\s*\\\\)$", "$1") as ?latitude)
    }
    """)
    same_as = Attribute("owl:sameAs")


class E53_PlaceFromGeonames:
    filter_type_is_feature = Filter([("rdf:type", "gn:Feature")])

    label = Attribute(["gn:name", "gn:officialName", "gn:alternateName"])
    latitude = Attribute("wgs84_pos:lat")
    longitude = Attribute("wgs84_pos:long")
    same_as = Attribute(["rdfs:seeAlso", "gn:wikipediaArticle"])


class E53_PlaceFromWikidata:
    filter_p625_must_exist = Filter([("wdt:P625", True)])

    label = Attribute(
        [
            "rdfs:label,de",
            "rdfs:label,en",
            "wdt:P1448/rdfs:label",
            "rdfs:label",
        ]
    )
    longitude = Attribute("""
    SELECT ?longitude
    WHERE {
      ?subject wdt:P625 ?geo1 .
      BIND(REPLACE(str(?geo1), "Point\\\\(([-+]?\\\\d+\\\\.?\\\\d+).*$", "$1") as ?longitude)
      }
    """)
    latitude = Attribute("""
    SELECT ?latitude
    WHERE {
      ?subject wdt:P625 ?geo1 .
      BIND(REPLACE(str(?geo1), "Point\\\\(([-+]?\\\\d+\\\\.?\\\\d+) ([-+]?\\\\d+\\\\.?\\\\d+).*$", "$2") as ?latitude)
      }
    """)
    same_as = Attribute(
        ["owl:sameAs", "wdtn:P227", "wdtn:P1566", "wdtn:P214", "wdtn:P244"]
    )



from django import forms

from apis_core.apis_entities.widgets import PlaceLookup


class PlaceLookupField(forms.CharField):
    widget = PlaceLookup

    def __init__(
        self,
        *args,
        label_id="id_label",
        longitude_id="id_longitude",
        latitude_id="id_latitude",
        **kwargs,
    ):
        self.label_id = label_id
        self.longitude_id = longitude_id
        self.latitude_id = latitude_id
        super().__init__(*args, **kwargs)

    def widget_attrs(self, widget):
        attrs = super().widget_attrs(widget)
        attrs["data-label-id"] = self.label_id
        attrs["data-longitude-id"] = self.longitude_id
        attrs["data-latitude-id"] = self.latitude_id
        return attrs
